/*
 * Copyright (C) 2001, Shilad Sen, Sourcelight Technologies, Inc.
 * See xmlrpc.h or the README for more copyright information.
 */


#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include "xmlrpc.h"
#include "rpcInternal.h"


#ifdef MSWINDOWS
	#include <winsock2.h>
#else
	#include <netdb.h>
	#include <netinet/in.h>
	#include <sys/socket.h>
	#include <unistd.h>
#endif



#define	READ_SIZE	4096


static	bool		connecting(
				rpcDisp		*dp,
				rpcSource	*sp,
				int		actions,
				PyObject	*params
			);
static	bool		writeRequest(
				rpcDisp		*dp,
				rpcSource	*sp,
				int		actions,
				PyObject	*params
			);
static	bool		readResponse(
				rpcDisp		*dp,
				rpcSource	*sp,
				int		actions,
				PyObject	*params
			);
static	bool		clientReadHeader(
				rpcDisp		*dp,
				rpcSource	*sp,
				int		actions,
				PyObject	*params
			);
static	bool		nbRead(int fd, PyObject **buffpp, bool *eof);
static	bool		executed(
				rpcDisp		*dp,
				rpcSource	*sp,
				int		actions,
				PyObject	*params
			);
static	PyObject	*pyRpcClientConnect(PyObject *self, PyObject *args);
static	PyObject	*pyRpcClientExecute(PyObject *self, PyObject *args);
static	PyObject	*pyRpcClientGetAttr(rpcClient *cp, char *name);
static	PyObject	*pyRpcClientWork(PyObject *self, PyObject *args);
static	bool		pyClientCallback(
			rpcDisp		*dp,
			rpcSource	*sp,
			int		actions,
			PyObject	*params
		);


rpcClient *
rpcClientNew(void)
{
	rpcClient	*cp;
	int		fd;
#ifdef MSWINDOWS
	ulong		flag	= 1;
#endif /* MSWINDOWS */

	cp = PyObject_NEW(rpcClient, &rpcClientType);
	if (cp == NULL)
		return NULL;
	cp->disp = dispNew();
	if (cp->disp == NULL)
		return NULL;
#ifdef MSWINDOWS
	fd = socket(PF_INET, SOCK_STREAM, 0);
	if (fd == INVALID_SOCKET)
		return (rpcClient *)PyErr_SetFromErrno(rpcError);
	unless (ioctlsocket((SOCKET)fd, FIONBIO, &flag) == 0)
		return (rpcClient *)PyErr_SetFromErrno(rpcError);
#else
	fd = socket(AF_INET, SOCK_STREAM, 0);
	if (fd < 0)
		return (rpcClient *)PyErr_SetFromErrno(rpcError);
	unless (fcntl(fd, F_SETFL, O_NONBLOCK) == 0)
		return (rpcClient *)PyErr_SetFromErrno(rpcError);
#endif /* MSWINDOWS */
	cp->src = rpcSourceNew(fd);
	if (cp->src == NULL)
		return NULL;
	cp->src->doClose = true;
	return  cp;
}


rpcClient *
rpcClientNewFromServer(rpcServer *servp)
{
	rpcClient	*cp;
	int		fd;
#ifdef MSWINDOWS
	ulong		flag	= 1;
#endif /* MSWINDOWS */

	cp = PyObject_NEW(rpcClient, &rpcClientType);
	if (cp == NULL)
		return NULL;
	cp->disp = servp->disp;
	Py_INCREF(servp->disp);
	fd = socket(AF_INET, SOCK_STREAM, 0);
#ifdef MSWINDOWS
	if (fd == INVALID_SOCKET)
		return (rpcClient *)PyErr_SetFromErrno(rpcError);
	unless (ioctlsocket((SOCKET)fd, FIONBIO, &flag) == 0)
		return (rpcClient *)PyErr_SetFromErrno(rpcError);
#else
	if (fd < 0)
		return (rpcClient *)PyErr_SetFromErrno(rpcError);
	unless (fcntl(fd, F_SETFL, O_NONBLOCK) == 0)
		return (rpcClient *)PyErr_SetFromErrno(rpcError);
#endif /* MSWINDOWS */
	cp->src = rpcSourceNew(fd);
	if (cp->src == NULL)
		return NULL;
	cp->src->doClose = true;
	return  cp;
}


void
rpcClientDealloc(rpcClient *cp)
{
	close(cp->src->fd);
	Py_DECREF(cp->src);
	Py_DECREF(cp->disp);
	PyMem_DEL(cp);
}


bool
rpcClientNbConnect(
	rpcClient	*cp,
	char		*host,
	int		port,
	bool		(*func) (rpcDisp *, rpcSource *, int, PyObject *),
	PyObject	*params
)
{
	rpcSource		*sp;
	struct sockaddr_in	addr;
	struct hostent		*hp;
	int			slen;

	sp = cp->src;
	hp = gethostbyname(host);
	if (hp == NULL) {
		PyErr_SetFromErrno(rpcError);
		return false;
	}
	addr.sin_family = hp->h_addrtype;
	memcpy(&addr.sin_addr, hp->h_addr, hp->h_length);
	addr.sin_port = htons((ushort) port);
	unless ((connect(sp->fd, (struct sockaddr *)&addr, sizeof(addr)) == 0)
	or      (isBlocked(get_errno()))) {
		PyErr_SetFromErrno(rpcError);
		return false;
	}
	slen = strlen(host) + strlen(":123456") + 1;	/* max desc length */
	sp->desc = alloc(slen);
	if (sp->desc == NULL)
		return false;
	snprintf(sp->desc, slen, "%s:%i", host, port);
	sp->desc[slen - 1] = EOS;
	sp->actImp = ACT_OUTPUT;
	sp->params = params;
	sp->func = func;
	unless (dispAddSource(cp->disp, sp))
		return false;

	return true;
}


bool
rpcClientConnect(rpcClient *cp, char *host, int port, double timeout)
{
	Py_INCREF(Py_None);
	unless ((rpcClientNbConnect(cp, host, port, connecting, Py_None))
	and     (dispWork(cp->disp, timeout)))
		return false;
	unless ((cp->src->params != NULL)
	and     (PyInt_Check(cp->src->params))) {
		set_errno(ETIMEDOUT);
		PyErr_SetFromErrno(rpcError);
		return false;
	}
	Py_DECREF(cp->src->params);
	cp->src->params = NULL;

	return true;
}


static bool
connecting(rpcDisp *dp, rpcSource *sp, int actions, PyObject *params)
{
	int	val,
		len;

	len = sizeof(val);
	if (getsockopt(sp->fd, SOL_SOCKET, SO_ERROR, (char *)&val, &len) < 0) {
		PyErr_SetFromErrno(rpcError);
		return false;
	}
	if (val == 0) {
		rpcLogSrc(1, sp, "client connection succeeded");
		sp->params = PyInt_FromLong(1);
		if (sp->params == NULL)
			return false;
		return true;
	} else unless (isBlocked(val)) {
		close(sp->fd);
		set_errno(val);
		PyErr_SetFromErrno(rpcError);
		return false;
	} else unless (dispAddSource(dp, sp))
		return false;
	rpcLogSrc(7, sp, "client connection poll gave %s", strerror(val));
	Py_INCREF(Py_None);
	sp->params = Py_None;
	sp->actImp = ACT_OUTPUT;
	sp->func = connecting;

	return true;
}


bool
rpcClientNbExecute(
	rpcClient	*cp,
	char		*method,
	PyObject	*args, 
	bool		(*func) (rpcDisp *, rpcSource *, int, PyObject *),
	PyObject	*params,
	PyObject	*addInfo
)
{
	PyObject	*req,
			*strReq;
	rpcSource	*sp;

	
	sp = cp->src;
	if (rpcLogLevel >= 5) {
		strReq = PyObject_Str(params);
		if (strReq == NULL)
			return false;
		rpcLogSrc(5, sp, "client queueing command ('%s', %s)",
			method, PyString_AS_STRING(strReq));
		Py_DECREF(strReq);
	} else if (rpcLogLevel >= 3)
		rpcLogSrc(3, sp, "client queueing command '%s'", method);

	req = buildRequest(method, params, addInfo);
	if (req == NULL)
		return false;
	if (rpcLogLevel >= 9) {
		strReq = PyObject_Repr(req);
		if (strReq == NULL)
			return false;
		rpcLogSrc(9, sp, "client request is %s",
				PyString_AS_STRING(strReq));
		Py_DECREF(strReq);
	}
	sp->actImp = ACT_OUTPUT;
	sp->func = writeRequest;
	sp->params = Py_BuildValue("(O,(s#,O))", req, &func, sizeof(func), args);
	Py_DECREF(req);
	unless ((sp->params != NULL)
	and     (dispAddSource(cp->disp, sp)))
		return false;
	return true;
}


PyObject *
rpcClientExecute(
	rpcClient	*cp,
	char		*method,
	PyObject	*params,
	PyObject	*addInfo,
	double		timeout
)
{
	PyObject	*result,
			*response;

	Py_INCREF(Py_None);
	unless ((rpcClientNbExecute(cp, method, Py_None, executed,
			params, addInfo))
	and     (dispWork(cp->disp, timeout)))
		return false;
	response = cp->src->params;
	cp->src->params = NULL;
	result = parseResponse(response);
	Py_DECREF(response);

	return result;
}


static bool
executed(rpcDisp *dp, rpcSource *sp, int actions, PyObject *params)
{
	PyObject	*res,
			*args;

	unless (PyArg_ParseTuple(params, "OO:writeSome", &res, &args))
		return false;

	sp->params = res;
	Py_INCREF(res);

	return true;
}


/*
 * write some of a request
 */
static bool
writeRequest(rpcDisp *dp, rpcSource *sp, int actions, PyObject *params)
{
	PyObject	*toWrite,
			*args;
	bool		res;
	int		nb,
			slen;

	unless (PyArg_ParseTuple(params, "SO:writeRequest", &toWrite, &args))
		return false;
	slen = PyString_GET_SIZE(toWrite);
	nb = write(sp->fd, PyString_AS_STRING(toWrite), slen);
	rpcLogSrc(7, sp, "client wrote %d of %d bytes", nb, slen);
	if (nb < 0 and isBlocked(get_errno()))
		nb = 0;
	if (nb < 0) {
		PyErr_SetFromErrno(rpcError);
		return false;
	} else if (nb == slen) {
		rpcLogSrc(7, sp, "client finished writing request");
		args = Py_BuildValue("(s,O)", "", args);
		if (args == NULL)
			return false;
		res = clientReadHeader(dp, sp, actions, args);
		Py_DECREF(args);
		return res;
	} else {
		assert(slen > nb);
		toWrite = PyString_FromStringAndSize(
			PyString_AS_STRING(toWrite) + nb, slen - nb);
		if (toWrite == NULL)
			return false;
		sp->actImp = ACT_OUTPUT;
		sp->func = writeRequest;
		sp->params = Py_BuildValue("(O,O)", toWrite, args);
		if (sp->params == NULL)
			return false;
		Py_DECREF(toWrite);
		unless (dispAddSource(dp, sp))
			return false;
		return true;
	}
}


static bool
clientReadHeader(rpcDisp *dp, rpcSource *sp, int actions, PyObject *params)
{
	PyObject	*buff,
			*args;
	bool		eof,
			res;
	ulong		blen;
	char		*hp,		/* start of header */
			*bp,		/* start of body */
			*cp,		/* current position */
			*ep,		/* end of string read */
			*lp;		/* start of content-length value */

	unless (PyArg_ParseTuple(params, "SO:clientReadHeader", &buff, &args))
		return false;
	unless (nbRead(sp->fd, &buff, &eof))
		return false;
	bp = NULL;
	lp = NULL;
	hp = PyString_AS_STRING(buff);
	ep = hp + PyString_GET_SIZE(buff);
	rpcLogSrc(9, sp, "client read %d bytes of header and body", (ep - hp));
	for (cp = hp; (bp == NULL) and (cp < ep); ++cp) {
		if ((ep - cp > 16)
		and (strncmp(cp, "Content-length: ", 16) == 0))
			lp = cp + 16;
		if ((ep - cp > 4)
		and (strncmp(cp, "\r\n\r\n", 4) == 0))
			bp = cp + 4;
		if ((ep - cp > 2)
		and (strncmp(cp, "\n\n", 2) == 0))
			bp = cp + 2;
	}
	if (bp == NULL) {
		if (eof) {
			Py_DECREF(buff);
			PyErr_SetString(rpcError, "got EOS while reading");
			return false;
		}
		sp->actImp = ACT_INPUT;
		sp->func = clientReadHeader;
		sp->params = Py_BuildValue("(O,O)", buff, args);
		Py_DECREF(buff);
		if ((sp->params == NULL)
		or  (not dispAddSource(dp, sp)))
			return false;
		return true;
	}
	if (lp == NULL) {
		Py_DECREF(buff);
		PyErr_SetString(rpcError,
			"no Content-length parameter found in header");
		return false;
	}
	unless (decodeActLong(&lp, ep, &blen)) {
		Py_DECREF(buff);
		PyErr_SetString(rpcError, "invalid Content-length");
		return false;
	}
	rpcLogSrc(9, sp, "client finished reading header");
	rpcLogSrc(9, sp, "client bodylen should be %d", blen);
	args = Py_BuildValue("(s#,s#,l,O)", hp, bp-hp, bp, ep-bp, blen, args);
	if (args == NULL)
		return false;
	res = readResponse(dp, sp, actions, args);
	Py_DECREF(args);
	Py_DECREF(buff);

	return res;
}


static bool
readResponse(rpcDisp *dp, rpcSource *sp, int actions, PyObject *params)
{
	PyObject	*head,
			*body,
			*args,
			*pyfunc;
	bool		(*func) (rpcDisp *, rpcSource *, int, PyObject *),
			eof,
			res;
	ulong		blen,
			slen;

	unless (PyArg_ParseTuple(params, "SSlO:readResponse",
					&head, &body, &blen, &args))
		return false;
	unless (nbRead(sp->fd, &body, &eof))
		return false;
	slen = PyString_GET_SIZE(body);
	rpcLogSrc(9, sp, "client read %d of %d bytes of body", slen, blen);
	func = NULL;				/* to appease the compiler */
	if (slen > blen) {
		Py_DECREF(body);
		PyErr_SetString(rpcError, "readResponse read too many bytes");
		return false;
	} else if (blen == slen) {
		PyObject	*ohead;
		ohead = head;
		rpcLogSrc(9, sp, "client finished reading body");
		Py_INCREF(head);	/* hack so concat doesn't fail */
		PyString_Concat(&head, body);
		unless ((head != NULL)
		and     (PyArg_ParseTuple(args, "SO:readResponse", &pyfunc, &args)))
			return false;
		Py_DECREF(body);
		if (rpcLogLevel >= 9) {
			PyObject *strReq;

			strReq = PyObject_Repr(head);
			if (strReq == NULL)
				return false;
			rpcLogSrc(9, sp, "client response was %s",
					PyString_AS_STRING(strReq));
			Py_DECREF(strReq);
		}
		args = Py_BuildValue("(O,O)", head, args);
		if (args == NULL)
			return false;
		memcpy(&func, PyString_AS_STRING(pyfunc), sizeof(func));
		res = func(dp, sp, actions, args);
		Py_DECREF(head);
		Py_DECREF(args);
		return res;
	} else {
		args = Py_BuildValue("(S,S,l,O)", head, body, blen, args);
		if (args == NULL)
			return false;
		Py_DECREF(body);
		sp->actImp = ACT_INPUT;
		sp->func = readResponse;
		sp->params = args;
		unless (dispAddSource(dp, sp))
			return false;
		return true;
	}
}


/*
 * returns number of bytes read, 
 */
static bool
nbRead(int fd, PyObject **buffpp, bool *eof)
{
	PyObject	*buffp;
	ulong		bytesAv,
			olen,
			slen;
	char		*cp;
	int		res;
	
	*eof = false;
	buffp = *buffpp;
	assert(PyString_Check(buffp));
	olen = PyString_GET_SIZE(buffp);
	slen = olen;
	bytesAv = slen + READ_SIZE;
	cp = alloc(bytesAv);
	if (cp == NULL)
		return false;
	memcpy(cp, PyString_AS_STRING(buffp), slen);
	while (true) {
		if (slen + READ_SIZE > bytesAv) {
			bytesAv = max(bytesAv * 2, slen + READ_SIZE);
			cp = ralloc(cp, bytesAv);
			if (cp == NULL)
				return false;
		}
		res = read(fd, cp + slen, READ_SIZE);
		if (res > 0)
			slen += res;
		else if (res == 0) {
			*eof = true;
			break;
		} else if (res < 0) {
			 if (isBlocked(get_errno()))
				break;
			else {		/* bad error */
				PyErr_SetFromErrno(rpcError);
				return false;
			}
		}
	}
	buffp = PyString_FromStringAndSize(cp, slen);
	if (buffp == NULL)
		return false;
	*buffpp = buffp;
	free(cp);

	return true;
}

/*
 * Module procedure: connect to an rpc Server
 */
static PyObject *
pyRpcClientConnect(PyObject *self, PyObject *args)
{
	char	*host;
	int	port;
	double	timeout;

	unless ((PyArg_ParseTuple(args, "sid", &host, &port, &timeout))
	and     (rpcClientConnect((rpcClient *)self, host, port, timeout)))
		return NULL;

	Py_INCREF(Py_None);
	return Py_None;
}


/*
 * Module procedure: execute a command on a rpc Server
 */
static PyObject *
pyRpcClientExecute(PyObject *self, PyObject *args)
{
	char		*method;
	PyObject	*params,
			*addInfo,
			*res;
	double		timeout;

	unless (PyArg_ParseTuple(args, "sOOd",
			&method, &params, &addInfo, &timeout))
		return NULL;
	unless (PySequence_Check(params)) {
		PyErr_SetString(rpcError, "execute params must be a sequence");
		return NULL;
	}
	unless (PyDict_Check(addInfo)) {
		PyErr_SetString(rpcError, "addInfo params must be a dictionary");
		return NULL;
	}
	res = rpcClientExecute((rpcClient *)self, method,
			params, addInfo, timeout);

	return res;
}


/*
 * queue up a function for later execution
 */
static PyObject *
pyRpcNbClientExecute(PyObject *self, PyObject *args)
{
	rpcClient	*cp;
	char		*method;
	PyObject	*params,
			*addInfo,
			*extArgs,
			*pyfunc;
	bool		res;

	cp = (rpcClient *)self;
	unless (PyArg_ParseTuple(args, "sOOOO",
			&method, &params, &addInfo, &pyfunc, &extArgs))
		return NULL;
	unless (PySequence_Check(params)) {
		PyErr_SetString(rpcError, "execute params must be a sequence");
		return NULL;
	}
	unless (PyDict_Check(addInfo)) {
		PyErr_SetString(rpcError, "addInfo params must be a dictionary");
		return NULL;
	}
	extArgs = Py_BuildValue("(O,O,O)", self, pyfunc, extArgs);
	if (params == NULL)
		return NULL;
	res = rpcClientNbExecute(cp, method, extArgs,
				pyClientCallback, params, addInfo);
	Py_DECREF(extArgs);
	unless (res)
		return NULL;

	Py_INCREF(Py_None);
	return Py_None;
}
	

static	bool
pyClientCallback(rpcDisp *dp, rpcSource *sp, int actions, PyObject *params)
{
	PyObject	*client,
			*pyfunc,
			*extArgs,
			*resp,
			*res;

	unless (PyArg_ParseTuple(params, "S(OOO):pyClientCallback",
					&resp, &client, &pyfunc, &extArgs))
		return false;

	assert(PyCallable_Check(pyfunc));
	params = Py_BuildValue("(O,O,O)", client, resp, extArgs);
	if (params == NULL)
		return false;
	res = PyObject_CallObject(pyfunc, params);
	Py_DECREF(params);
	unless (res)
		return false;

	return true;
}


/*
 * Work on a socket for a while
 */
static PyObject *
pyRpcClientWork(PyObject *self, PyObject *args)
{
	double		timeout;
	rpcClient	*cp;

	cp = (rpcClient *)self;
	unless ((PyArg_ParseTuple(args, "d", &timeout))
	and     (dispWork(cp->disp, timeout)))
		return NULL;

	Py_INCREF(Py_None);
	return Py_None;
}


/*
 * Set a handler for errors on the client
 */
static PyObject *
pyRpcClientSetOnErr(PyObject *self, PyObject *args)
{
	PyObject	*func;
	rpcClient	*cp;

	cp = (rpcClient *)self;
	unless (PyArg_ParseTuple(args, "O", &func))
		return NULL;
	unless (PyCallable_Check(func)) {
		PyErr_SetString(rpcError, "error handler must be callable");
		return NULL;
	}
	if (PyObject_Compare(func, Py_None))
		rpcSourceSetOnErr(cp->src, ONERR_TYPE_PY, func);
	else
		rpcSourceSetOnErr(cp->src, ONERR_TYPE_DEF, NULL);

	Py_INCREF(Py_None);
	return Py_None;
}


/*
 * member functions for client object
 */
static PyMethodDef pyRpcClientMethods[] = {
	{ "work",	(PyCFunction)pyRpcClientWork,		1,	0 },
	{ "connect",	(PyCFunction)pyRpcClientConnect,	1,	0 },
	{ "execute",	(PyCFunction)pyRpcClientExecute,	1,	0 },
	{ "nbexecute",	(PyCFunction)pyRpcNbClientExecute,	1,	0 },
	{ "setOnErr",	(PyCFunction)pyRpcClientSetOnErr,	1,	0 },
	{ NULL,		NULL},
};


/*
 * return an attribute for a client object
 */
static PyObject *
pyRpcClientGetAttr(rpcClient *cp, char *name)
{
	return Py_FindMethod(pyRpcClientMethods, (PyObject *)cp, name);
}


/*
 * map characterstics of a client object
 */
PyTypeObject rpcClientType = {
	PyObject_HEAD_INIT(0)
	0,
	"rpcClient",
	sizeof(rpcClient),
	0,
	(destructor)rpcClientDealloc,		/* tp_dealloc */
	0,					/* tp_print */
	(getattrfunc)pyRpcClientGetAttr,	/* tp_getattr */
	0,					/* tp_setattr */
	0,					/* tp_compare */
	0,					/* tp_repr */
	0,					/* tp_as_number */
	0,					/* tp_as_sequence */
	0,					/* tp_as_mapping */
	0,					/* tp_hash */
	0,					/* tp_call */
	0,					/* tp_str */
	0,					/* tp_getattro */
	0,					/* tp_setattro */
	0,					/* tp_as_buffer */
	0,					/* tp_xxx4 */
	0,					/* tp_doc */
};
