/*
 * Copyright (C) 2001, Shilad Sen, Sourcelight Technologies, Inc.
 * See xmlrpc.h or the README for more copyright information.
 */


#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include "xmlrpc.h"
#include "rpcInternal.h"


#ifdef MSWINDOWS
	#include <winsock2.h>
#else
	#include <sys/socket.h>
	#include <netinet/in.h>
	#include <unistd.h>
#endif


#define	READ_SIZE	4096


static	bool		serveAccept(
				rpcDisp		*dp,
				rpcSource	*sp,
				int		actions,
				PyObject	*params
			);
static	bool		serverReadHeader(
				rpcDisp		*dp,
				rpcSource	*sp,
				int		actions,
				PyObject	*params
			);
static	bool		readRequest(
				rpcDisp		*dp,
				rpcSource	*sp,
				int		actions,
				PyObject	*params
			);
static	bool		writeResponse(
				rpcDisp		*dp,
				rpcSource	*sp,
				int		actions,
				PyObject	*params
			);
static	bool		doResponse(
				rpcServer	*servp,
				rpcSource	*srcp,
				PyObject	*request
			);
static	PyObject	*dispatch(
				rpcServer	*servp,
				rpcSource	*srcp,
				PyObject	*request
			);
static	bool		nbRead(int fd, PyObject **buffpp, bool *eof);
static	char		*pyerrToStr(void);


rpcServer *
rpcServerNew(void)
{
	rpcServer	*sp;
	int		fd;
#ifdef MSWINDOWS
	ulong		flag	= 1;
#endif /* MSWINDOWS */

	sp = PyObject_NEW(rpcServer, &rpcServerType);
	if (sp == NULL)
		return NULL;
	sp->keepAlive = false;
	sp->disp = dispNew();
	if (sp->disp == NULL)
		return NULL;
#ifdef MSWINDOWS
	fd = socket(PF_INET, SOCK_STREAM, 0);
	if (fd == INVALID_SOCKET)
		return (rpcServer *)PyErr_SetFromErrno(rpcError);
	unless (ioctlsocket((SOCKET)fd, FIONBIO, &flag) == 0)
		return (rpcServer *)PyErr_SetFromErrno(rpcError);
#else
	fd = socket(AF_INET, SOCK_STREAM, 0);
	if (fd < 0)
		return (rpcServer *)PyErr_SetFromErrno(rpcError);
	unless (fcntl(fd, F_SETFL, O_NONBLOCK) == 0)
		return (rpcServer *)PyErr_SetFromErrno(rpcError);
#endif /* MSWINDOWS */
	sp->src = rpcSourceNew(fd);
	if (sp->src == NULL)
		return NULL;
	sp->src->doClose = true;
	sp->comtab = PyDict_New();
	if (sp->comtab == NULL)
		return NULL;
	return  sp;
}


void
rpcServerDealloc(rpcServer *sp)
{
	Py_DECREF(sp->src);
	dispDealloc(sp->disp);
}


bool
rpcServerAddCMethod(
	rpcServer 	*servp,
	char		*method,
	PyObject	*(*cfunc)(
				rpcServer *,
				rpcSource *,
				char *,
				PyObject *,
				PyObject *
			)
)
{
	PyObject	*pyfunc;

	pyfunc = PyString_FromStringAndSize((char*)&cfunc, sizeof(cfunc));
	if (pyfunc == NULL)
		return false;
	if (PyDict_SetItemString(servp->comtab, method, pyfunc))
		return false;
	return true;
}


bool
rpcServerAddPyMethods(rpcServer *sp, PyObject *toAdd)
{
	PyObject	*items,
			*elem,
			*method,
			*func;
	int		i;
	
	unless (PyDict_Check(toAdd)) {
		PyErr_SetString(rpcError, "addMethods requires dictionary");
		return false;
	}
	items = PyDict_Items(toAdd);
	if (items == NULL)
		return false;
	for (i = 0; i < PyList_GET_SIZE(items); ++i) {
		elem = PyList_GET_ITEM(items, i);
		assert(PyTuple_Check(elem));
		assert(PyTuple_GET_SIZE(elem) == 2);
		method = PyTuple_GET_ITEM(elem, 0);
		func = PyTuple_GET_ITEM(elem, 1);
		unless (PyString_Check(method)) {
			PyErr_SetString(rpcError, "method names must be strings");
			return false;
		}
		unless (PyCallable_Check(func)) {
			PyErr_SetString(rpcError, "method must be callable");
			return false;
		}
		if (PyDict_SetItem(sp->comtab, method, func))
			return false;
	}
	return true;
}


bool 
rpcServerBindAndListen(rpcServer *servp, int port, int queue)
{
	int			fd,
				sflag;
	struct sockaddr_in	saddr;

	sflag = 1;
	fd = servp->src->fd;
	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
		(void *)&sflag, sizeof(sflag)) != 0) {
		PyErr_SetFromErrno(rpcError);
		return false;
	}
	saddr.sin_family = AF_INET;
	saddr.sin_addr.s_addr = htonl(INADDR_ANY);
	saddr.sin_port = htons((ushort) port);
	if ((bind(fd, (struct sockaddr *)&saddr, sizeof(saddr)) < 0)
	or  (listen(fd, queue) < 0)) {
		PyErr_SetFromErrno(rpcError);
		return false;
	}
	rpcLogSrc(3, servp->src, "server listening on port %d", port);
	Py_INCREF(servp);
	servp->src->actImp = ACT_INPUT;
	servp->src->func = serveAccept;
	servp->src->params = (PyObject *)servp;
	unless (dispAddSource(servp->disp, servp->src))
		return false;
	return true;
}


bool 
rpcServerSetFdAndListen(rpcServer *servp, int fd, int queue)
{
#ifdef MSWINDOWS
	ulong		flag	= 1;

	unless ((ioctlsocket((SOCKET)fd, FIONBIO, &flag) == 0)
	and     (listen(fd, queue) >= 0)) {
			PyErr_SetFromErrno(rpcError);
			return false;
		}
#else
	unless ((fcntl(fd, F_SETFL, O_NONBLOCK) == 0)
	and     (listen(fd, queue) >= 0)) {
			PyErr_SetFromErrno(rpcError);
			return false;
		}
#endif /* MSWINDOWS */
	rpcLogSrc(3, servp->src, "server listening on fd %d", fd);
	Py_INCREF(servp);
	servp->src->fd = fd;
	servp->src->actImp = ACT_INPUT;
	servp->src->func = serveAccept;
	servp->src->params = (PyObject *)servp;
	unless (dispAddSource(servp->disp, servp->src))
		return false;
	return true;
}


static bool
serveAccept(rpcDisp *dp, rpcSource *sp, int actions, PyObject *servp)
{
	int			len,
				res;
	uint			in;
	rpcSource		*client;
	struct sockaddr_in	addr;
#ifdef MSWINDOWS
	ulong		flag	= 1;
#endif /* MSWINDOWS */

	len = sizeof(addr);
	res = accept(sp->fd, (struct sockaddr *)&addr, &len);
	if (res >= 0) {
#ifdef MSWINDOWS
		unless (ioctlsocket((SOCKET)res, FIONBIO, &flag) == 0) {
			PyErr_SetFromErrno(rpcError);
			return false;
		}
#else
		unless (fcntl(res, F_SETFL, O_NONBLOCK) == 0) {
			PyErr_SetFromErrno(rpcError);
			return false;
		}
#endif /* MSWINDOWS */
		client = rpcSourceNew(res);
		if (client == NULL)
			return false;
		client->doClose = true;
		client->desc = alloc(strlen("255.255.255.255:123456" + 1));
		if (client->desc == NULL)
			return false;
		in = ntohl(addr.sin_addr.s_addr);
		sprintf(client->desc, "%u.%u.%u.%u:%u",
			0xFF & (in>>24), 0xFF & (in>>16),
			0xFF & (in>>8), 0xFF & in, ntohs(addr.sin_port));
		rpcLogSrc(3, sp, "server got connection from %s", client->desc);
		client->actImp = ACT_INPUT;
		client->func = serverReadHeader;
		client->params = Py_BuildValue("(s,O)", "", servp);
		if (client->params == NULL)
			return false;
		unless (dispAddSource(dp, client))
			return false;
		Py_DECREF(client);
	} else unless (isBlocked(get_errno())) {
		PyErr_SetFromErrno(rpcError);
		return false;
	} else
		fprintf(stderr, "blocked on accept\n");
	sp->actImp = ACT_INPUT;
	sp->func = serveAccept;
	sp->params = servp;
	Py_INCREF(servp);
	unless (dispAddSource(dp, sp))
		return false;
	return true;
}



static bool
serverReadHeader(rpcDisp *dp, rpcSource *sp, int actions, PyObject *params)
{
	PyObject	*buff,
			*args,
			*servp;
	bool		eof,
			res;
	ulong		blen;
	char		*hp,		/* start of header */
			*bp,		/* start of body */
			*cp,		/* current position */
			*ep,		/* end of string read */
			*lp;		/* start of content-length value */

	unless (PyArg_ParseTuple(params, "SO:serverReadHeader", &buff, &servp))
		return false;
	eof = false;
	unless (nbRead(sp->fd, &buff, &eof))
		return false;
	bp = NULL;
	lp = NULL;
	hp = PyString_AS_STRING(buff);
	ep = hp + PyString_GET_SIZE(buff);
	rpcLogSrc(7, sp, "server read %d bytes of header",
			PyString_GET_SIZE(buff));
	for (cp = hp; (bp == NULL) and (cp < ep); ++cp) {
		if ((ep - cp > 16)
		and (strncmp(cp, "Content-length: ", 16) == 0))
			lp = cp + 16;
		else if ((ep - cp > 16)
		and      (strncmp(cp, "Content-Length: ", 16) == 0))
			lp = cp + 16;
		if ((ep - cp > 4)
		and (strncmp(cp, "\r\n\r\n", 4) == 0))
			bp = cp + 4;
		if ((ep - cp > 2)
		and (strncmp(cp, "\n\n", 2) == 0))
			bp = cp + 2;
	}
	if (bp == NULL) {
		if (eof) {
			if (PyString_GET_SIZE(buff) == 0) {
				Py_DECREF(buff);
				rpcLogSrc(3, sp, "received EOF");
				return true;
			}
			Py_DECREF(buff);
			PyErr_SetString(rpcError, "got EOS while reading");
			return false;
		}
		sp->actImp = ACT_INPUT;
		sp->func = serverReadHeader;
		sp->params = Py_BuildValue("(O,O)", buff, servp);
		Py_DECREF(buff);
		if ((sp->params == NULL)
		or  (not dispAddSource(dp, sp)))
			return false;
		return true;
	}
	if (lp == NULL) {
		Py_DECREF(buff);
		PyErr_SetString(rpcError,
			"no Content-length parameter found in header");
		return false;
	}
	unless (decodeActLong(&lp, ep, &blen)) {
		Py_DECREF(buff);
		PyErr_SetString(rpcError, "invalid Content-length");
		return false;
	}
	rpcLogSrc(7, sp, "server finished reading header");
	rpcLogSrc(9, sp, "server content length should be %d", blen);
	args = Py_BuildValue("(s#,s#,l,O)", hp, bp-hp, bp, ep-bp, blen, servp);
	if (args == NULL)
		return false;
	res = readRequest(dp, sp, actions, args);
	Py_DECREF(args);
	Py_DECREF(buff);

	return res;
}



static bool
readRequest(rpcDisp *dp, rpcSource *sp, int actions, PyObject *params)
{
	PyObject	*head,
			*body,
			*servp;
	bool		eof,
			res;
	ulong		blen,
			slen;

	unless (PyArg_ParseTuple(params, "SSlO:readRequest",
					&head, &body, &blen, &servp))
		return false;
	unless (nbRead(sp->fd, &body, &eof))
		return false;
	slen = PyString_GET_SIZE(body);
	rpcLogSrc(9, sp, "server read %d of %d body bytes", slen, blen);
	if (slen > blen) {
		Py_DECREF(body);
		PyErr_SetString(rpcError, "readRequest read too many bytes");
		return false;
	} else if (blen == slen) {
		rpcLogSrc(9, sp, "server finished reading body");
		Py_INCREF(head);	/* hack so concat doesn't fail */
		PyString_ConcatAndDel(&head, body);
		if (head == NULL)
			return false;
		res = doResponse((rpcServer *)servp, sp, head);
		Py_DECREF(head);
		return res;
	} else {
		if (eof) {
			Py_DECREF(body);
			PyErr_SetString(rpcError, "got EOS while reading body");
			return false;
		}
		sp->actImp = ACT_INPUT;
		sp->func = readRequest;
		sp->params = Py_BuildValue("(S,S,l,O)",
				head, body, blen, servp);
		Py_DECREF(body);
		if (sp->params == NULL)
			return false;
		unless (dispAddSource(dp, sp))
			return false;
		return true;
	}
}


static bool
doResponse(rpcServer *servp, rpcSource *srcp, PyObject *request)
{
	PyObject	*addInfo,
			*response,
			*result,
			*strRes,
			*params;
	char		*cerr;
	bool		res;

	addInfo = PyDict_New();
	if (addInfo == NULL)
		return false;
	result = dispatch(servp, srcp, request);
	response = NULL;
	if (result == NULL) {
		cerr = pyerrToStr();
		if (cerr == NULL)
			response = buildFault(1, "Unknown error", addInfo);
		else {
			response = buildFault(1, cerr, addInfo);
			free(cerr);
		}
	} else {
		response = buildResponse(result, addInfo);
		Py_DECREF(result);
	}
	/* one more attempt at failure recovery */
	if (response == NULL)
		response = buildFault(1, "Unknown error", addInfo);
	Py_DECREF(addInfo);
	if (response == NULL)
		return false;
	if (rpcLogLevel >= 8) {
		strRes = PyObject_Repr(response);
		if (strRes == NULL)
			return false;
		rpcLogSrc(8, srcp, "server responding with %s",
			PyString_AS_STRING(strRes));
		Py_DECREF(strRes);
	}
	params = Py_BuildValue("(O,O)", response, servp);
	Py_DECREF(response);
	if (params == NULL)
		return false;
	res = writeResponse(servp->disp, srcp, ACT_OUTPUT, params);
	Py_DECREF(params);

	return res;
}


static PyObject *
dispatch(rpcServer *servp, rpcSource *srcp, PyObject *request)
{
	PyObject	*args,
			*method,
			*params,
			*addInfo,
			*tuple,	
			*pyfunc,
			*(*cfunc)(
				rpcServer *,
				rpcSource *,
				char *,
				PyObject *,
				PyObject *
			),
			*result,
			*strReq,
			*strRes;
	char		buff[256];

	if (rpcLogLevel >= 8) {
		strReq = PyObject_Repr(request);
		if (strReq == NULL)
			return NULL;
		rpcLogSrc(8, srcp, "server got request %s",
			PyString_AS_STRING(strReq));
		Py_DECREF(strReq);
	}
	tuple = parseRequest(request);
	if (tuple == NULL)
		return NULL;
	assert(PyTuple_Check(tuple));
	assert(PyTuple_GET_SIZE(tuple) == 3);
	method = PyTuple_GET_ITEM(tuple, 0);
	params = PyTuple_GET_ITEM(tuple, 1);
	addInfo = PyTuple_GET_ITEM(tuple, 2);

	if (rpcLogLevel >= 5) {
		strReq = PyObject_Repr(params);
		if (strReq == NULL)
			return false;
		rpcLogSrc(5, srcp, "server got request ('%s', %s)",
			PyString_AS_STRING(method), PyString_AS_STRING(strReq));
		Py_DECREF(strReq);
	} else if (rpcLogLevel >= 3)
		rpcLogSrc(3, srcp, "server got request '%s'",
			PyString_AS_STRING(method));

	assert(PyString_Check(method));
	assert(PyDict_Check(addInfo));	/* fix: decrement ref?? */
	unless (PyMapping_HasKey(servp->comtab, method)) {
		snprintf(buff, 255, "unknown command: \'%s\'",
			PyString_AS_STRING(method));
		PyErr_SetString(rpcError, buff);
		return NULL;
	}
	pyfunc = PyDict_GetItem(servp->comtab, method);
	if (PyCallable_Check(pyfunc)) {
		args = Py_BuildValue("(O,O,O,O,O)",
				 servp, srcp, method, params, addInfo);
		Py_DECREF(tuple);
		if (args == NULL)
			return NULL;
		result = PyObject_CallObject(pyfunc, args);
		Py_DECREF(args);
	} else if (PyString_Check(pyfunc)) {
		assert(PyString_GET_SIZE(pyfunc) == sizeof(cfunc));
		memcpy(&cfunc, PyString_AS_STRING(pyfunc), sizeof(cfunc));
		result = cfunc(
				servp,
				srcp,
				PyString_AS_STRING(method),
				params,
				addInfo
			);
		Py_DECREF(tuple);
	} else {
		setPyErr("illegal type for server callback");
		return NULL;
	}
	if (result == NULL)
		return NULL;
	if (rpcLogLevel >= 5) {
		strRes = PyObject_Str(result);
		if (strRes == NULL)
			return false;
		rpcLogSrc(5, srcp, "server responding %s", 
			PyString_AS_STRING(strRes));
		Py_DECREF(strRes);
	}

	return result;
}


static bool
writeResponse(rpcDisp *dp, rpcSource *srcp, int actions, PyObject *params)
{
	rpcServer	*servp;
	PyObject	*toWrite;
	int		nb,
			slen;

	unless (PyArg_ParseTuple(params, "SO:writeResponse", &toWrite, &servp))
		return false;
	slen = PyString_GET_SIZE(toWrite);
	nb = write(srcp->fd, PyString_AS_STRING(toWrite), slen);
	rpcLogSrc(9, srcp, "server wrote %d of %d bytes", nb, slen);
	if (nb < 0 and isBlocked(get_errno()))
		nb = 0;
	if (nb < 0) {
		PyErr_SetFromErrno(rpcError);
		return false;
	} else if (nb == slen) {
		rpcLogSrc(9, srcp, "server finished writing response");
		srcp->actImp = ACT_INPUT;
		srcp->func = serverReadHeader;
		srcp->params = Py_BuildValue("(s,O)", "", servp);
		if (srcp->params == NULL)
			return false;
		if (servp->keepAlive) {
			unless (dispAddSource(dp, srcp))
				return false;
		} else
			close(srcp->fd);
		return true;
	} else {
		assert(slen > nb);
		toWrite = PyString_FromStringAndSize(
			PyString_AS_STRING(toWrite) + nb, slen - nb);
		if (toWrite == NULL)
			return false;
		srcp->actImp = ACT_OUTPUT;
		srcp->func = writeResponse;
		srcp->params = Py_BuildValue("(O,O)", toWrite, servp);
		Py_DECREF(toWrite);
		if (srcp->params == NULL)
			return false;
		unless (dispAddSource(dp, srcp))
			return false;
		return true;
	}
}


static char *
pyerrToStr(void)
{
	PyObject 	*exc;
	PyObject 	*v;
	PyObject 	*tb;
	PyObject	*err1;
	PyObject	*err2;
	PyObject	*delim;
	char		*cerr;

	unless (PyErr_Occurred())
		return NULL;
	PyErr_Fetch(&exc, &v, &tb);
	PyErr_NormalizeException(&exc, &v, &tb);
	if (exc == NULL)
		return NULL;
	err1 = PyObject_Str(exc);
	err2 = PyObject_Str(v);
	delim = PyString_FromString(": ");
	if (err1 == NULL || err2 == NULL || delim == NULL)
		return NULL;
	PyString_Concat(&err1, delim);
	if (err1 == NULL)
		return NULL;
	PyString_Concat(&err1, err2);
	if (err1 == NULL)
		return NULL;
	cerr = alloc(PyString_GET_SIZE(err1) + 1);
	if (cerr == NULL)
		return NULL;
	strcpy(cerr, PyString_AS_STRING(err1));
	PyErr_Restore(exc, v, tb);
	PyErr_Print();
	PyErr_Clear();
	Py_DECREF(delim);
	Py_DECREF(err1);
	Py_DECREF(err2);

	return cerr;
}


/*
 * Register some commands with an rpc server
 */
static PyObject *
pyRpcServerAddMethods(PyObject *self, PyObject *args)
{
	rpcServer	*sp;
	PyObject	*toAdd;

	sp = (rpcServer *)self;
	unless ((PyArg_ParseTuple(args, "O", &toAdd)
	and     (rpcServerAddPyMethods(sp, toAdd))))
		return NULL;

	Py_INCREF(Py_None);
	return Py_None;
}


/*
 * Bind an rpc server to a port, and start it listening
 */
static PyObject *
pyRpcServerBindAndListen(PyObject *self, PyObject *args)
{
	rpcServer	*sp;
	int		port,
			queue;

	sp = (rpcServer *)self;
	unless ((PyArg_ParseTuple(args, "ii", &port, &queue)
	and     (rpcServerBindAndListen(sp, port, queue))))
		return NULL;

	Py_INCREF(Py_None);
	return Py_None;
}


/*
 * Bind an rpc server to a port, and start it listening
 */
static PyObject *
pyRpcServerSetFdAndListen(PyObject *self, PyObject *args)
{
	rpcServer	*sp;
	int		fd,
			queue;

	sp = (rpcServer *)self;
	unless ((PyArg_ParseTuple(args, "ii", &fd, &queue)
	and     (rpcServerSetFdAndListen(sp, fd, queue))))
		return NULL;

	Py_INCREF(Py_None);
	return Py_None;
}


/*
 * Bind an rpc server to a port, and start it listening
 */
static PyObject *
pyRpcServerWork(PyObject *self, PyObject *args)
{
	double		timeout;
	rpcServer	*sp;

	sp = (rpcServer *)self;
	unless ((PyArg_ParseTuple(args, "d", &timeout))
	and     (dispWork(sp->disp, timeout)))
		return NULL;

	Py_INCREF(Py_None);
	return Py_None;
}


/*
 * Bind an rpc server to a port, and start it listening
 */
static PyObject *
pyRpcServerActiveFds(PyObject *self, PyObject *args)
{
	rpcServer	*sp;

	sp = (rpcServer *)self;
	unless (PyArg_ParseTuple(args, ""))
		return NULL;

	return dispActiveFds(sp->disp);
}


/*
 * Set a handler for errors on the client
 */
static PyObject *
pyRpcServerSetOnErr(PyObject *self, PyObject *args)
{
	PyObject	*func;
	rpcServer	*servp;

	servp = (rpcServer *)self;
	unless (PyArg_ParseTuple(args, "O", &func))
		return NULL;
	unless (PyCallable_Check(func)) {
		PyErr_SetString(rpcError, "error handler must be callable");
		return NULL;
	}
	if (PyObject_Compare(func, Py_None))
		rpcSourceSetOnErr(servp->src, ONERR_TYPE_PY, func);
	else
		rpcSourceSetOnErr(servp->src, ONERR_TYPE_DEF, NULL);

	Py_INCREF(Py_None);
	return Py_None;
}



/*
 * returns number of bytes read, 
 */
static bool
nbRead(int fd, PyObject **buffpp, bool *eof)
{
	PyObject	*buffp;
	ulong		bytesAv,
			olen,
			slen;
	char		*cp;
	int		res;
	
	*eof = false;
	buffp = *buffpp;
	assert(PyString_Check(buffp));
	olen = PyString_GET_SIZE(buffp);
	slen = olen;
	bytesAv = slen + READ_SIZE;
	cp = alloc(bytesAv);
	if (cp == NULL)
		return false;
	memcpy(cp, PyString_AS_STRING(buffp), slen);
	while (true) {
		if (slen + READ_SIZE > bytesAv) {
			bytesAv = max(bytesAv * 2, slen + READ_SIZE);
			cp = ralloc(cp, bytesAv);
			if (cp == NULL)
				return false;
		}
		res = read(fd, cp + slen, READ_SIZE);
		if (res > 0)
			slen += res;
		else if (res == 0) {
			*eof = true;
			break;
		} else if (res < 0) {
			 if (isBlocked(get_errno()))
				break;
			else {		/* bad error */
				PyErr_SetFromErrno(rpcError);
				return false;
			}
		}
	}
	buffp = PyString_FromStringAndSize(cp, slen);
	if (buffp == NULL)
		return false;
	*buffpp = buffp;
	free(cp);

	return true;
}


/*
 * Tell an rpc server to exit the "work routine" asap
 */
static PyObject *
pyRpcServerExit(PyObject *self)
{
	rpcServer	*servp;

	servp = (rpcServer *)self;
	servp->disp->etime = 0.0;
	Py_INCREF(Py_None);

	return Py_None;
}


/*
 * member functions for server object
 */
static PyMethodDef pyRpcServerMethods[] = {
	{ "addMethods",	    (PyCFunction)pyRpcServerAddMethods,     1, 0 },
	{ "bindAndListen",  (PyCFunction)pyRpcServerBindAndListen,  1, 0 },
	{ "setFdAndListen", (PyCFunction)pyRpcServerSetFdAndListen, 1, 0 },
	{ "activeFds",      (PyCFunction)pyRpcServerActiveFds,      1, 0 },
	{ "work",           (PyCFunction)pyRpcServerWork,           1, 0 },
	{ "exit",           (PyCFunction)pyRpcServerExit,           1, 0 },
	{ "setOnErr",       (PyCFunction)pyRpcServerSetOnErr,       1, 0 },
	{ NULL,		NULL},
};


/*
 * return an attribute for a server object
 */
static PyObject *
pyRpcServerGetAttr(rpcServer *sp, char *name)
{
	return Py_FindMethod(pyRpcServerMethods, (PyObject *)sp, name);
}


/*
 * map characterstics of an edb object
 */
PyTypeObject rpcServerType = {
	PyObject_HEAD_INIT(0)
	0,
	"rpcServer",
	sizeof(rpcServer),
	0,
	(destructor)rpcServerDealloc,		/* tp_dealloc */
	0,					/* tp_print */
	(getattrfunc)pyRpcServerGetAttr,	/* tp_getattr */
	0,					/* tp_setattr */
	0,					/* tp_compare */
	0,					/* tp_repr */
	0,					/* tp_as_number */
	0,					/* tp_as_sequence */
	0,					/* tp_as_mapping */
	0,					/* tp_hash */
	0,					/* tp_call */
	0,					/* tp_str */
	0,					/* tp_getattro */
	0,					/* tp_setattro */
	0,					/* tp_as_buffer */
	0,					/* tp_xxx4 */
	0,					/* tp_doc */
};
